@@ -18,74 +18,103 @@ from abc import abstractmethod
 import six
 
 from neutron_lib.api import converters as lib_converters
+from neutron_lib.api import extensions
+from neutron_lib.api import validators as lib_validators
+from neutron_lib.db import constants as db_const
 from neutron_lib import exceptions as neutron_exc
+from neutron_lib.services import base as service_base
 from oslo_config import cfg
 
 from neutron.api import extensions as neutron_ext
-from neutron.api.v2 import attributes as attr
 from neutron.api.v2 import resource_helper
-from neutron.services import service_base
 
 from networking_sfc._i18n import _
-from networking_sfc import extensions
+from networking_sfc import extensions as sfc_extensions
+from networking_sfc.extensions import flowclassifier as ext_fc
+
 
 cfg.CONF.import_opt('api_extensions_path', 'neutron.common.config')
-neutron_ext.append_api_extensions_path(extensions.__path__)
+neutron_ext.append_api_extensions_path(sfc_extensions.__path__)
 
 SFC_EXT = "sfc"
 SFC_PREFIX = "/sfc"
 
+# Default Chain Parameters
+DEFAULT_CHAIN_CORRELATION = 'mpls'
+DEFAULT_CHAIN_SYMMETRY = False
+DEFAULT_CHAIN_PARAMETERS = {'correlation': DEFAULT_CHAIN_CORRELATION,
+                            'symmetric': DEFAULT_CHAIN_SYMMETRY}
+
+# Default SF Parameters
+DEFAULT_SF_PARAMETERS = {'correlation': None, 'weight': 1}
+
+# Default and Supported PPG Parameters
+DEFAULT_PPG_LB_FIELDS = []
+DEFAULT_PPG_N_TUPLE = {'ingress_n_tuple': {}, 'egress_n_tuple': {}}
+DEFAULT_PPG_PARAMETERS = {'lb_fields': DEFAULT_PPG_LB_FIELDS,
+                          'ppg_n_tuple_mapping': DEFAULT_PPG_N_TUPLE}
 SUPPORTED_LB_FIELDS = [
     "eth_src", "eth_dst", "ip_src", "ip_dst",
     "tcp_src", "tcp_dst", "udp_src", "udp_dst"
 ]
-
-
-class InvalidLBField(neutron_exc.InvalidInput):
-    message = _("Unknown lb field %(field)s.")
-
-
-def normalize_lb_fields(lb_fields):
-    lb_fields = lib_converters.convert_none_to_empty_list(lb_fields)
-    for field in lb_fields:
-        if field not in SUPPORTED_LB_FIELDS:
-            raise InvalidLBField(
-                field=field)
-    return lb_fields
-
-
-SUPPORTED_CHAIN_PARAMETERS = {
-    'correlation': {
-        'allow_post': True,
-        'default': 'mpls',
-        'validate': {'type:values': ['mpls']}
-    }
-}
-SUPPORTED_SF_PARAMETERS = {
-    'correlation': {
-        'allow_post': True,
-        'default': None,
-        'validate': {'type:values': [None]}
-    },
-    'weight': {
-        'allow_post': True,
-        'default': 1,
-        'validate': {'type:non_negative': None},
-        'convert_to': lib_converters.convert_to_int
-    }
-}
-SUPPORTED_PPG_PARAMETERS = {
-    'lb_fields': {
-        'allow_post': True,
-        'default': None,
-        'validate': {'type:list_of_unique_strings': None},
-        'convert_to': normalize_lb_fields
-    }
+SUPPORTED_PPG_TUPLE_MAPPING = {
+    'source_ip_prefix': None,
+    'destination_ip_prefix': None,
+    'source_port_range_min': None,
+    'source_port_range_max': None,
+    'destination_port_range_min': None,
+    'destination_port_range_max': None,
 }
+
 MAX_CHAIN_ID = 65535
 
 
-# Port Chain Exceptions
+# NOTE(scsnow): move to neutron-lib
+def validate_list_of_allowed_values(data, allowed_values=None):
+    if not isinstance(data, list):
+        msg = _("'%s' is not a list") % data
+        return msg
+
+    illegal_values = set(data) - set(allowed_values)
+    if illegal_values:
+        msg = _("Illegal values in a list: %s") % ', '.join(illegal_values)
+        return msg
+
+
+lib_validators.validators['type:list_of_allowed_values'] = \
+    validate_list_of_allowed_values
+
+
+# DEFAULT RESOURCE_ATTRIBUTE_MAP for ingress_n_tuple and egress_n_tuple in
+# ppg_n_tuple_mapping validate dict
+ppg_n_tuple_validact_dict = {
+    'source_ip_prefix': {
+        'default': None,
+        'validate': {'type:subnet_or_none': None}
+    },
+    'destination_ip_prefix': {
+        'default': None,
+        'validate': {'type:subnet_or_none': None}
+    },
+    'source_port_range_min': {
+        'default': None,
+        'convert_to': ext_fc.normalize_port_value
+    },
+    'source_port_range_max': {
+        'default': None,
+        'convert_to': ext_fc.normalize_port_value
+    },
+    'destination_port_range_min': {
+        'default': None,
+        'convert_to': ext_fc.normalize_port_value
+    },
+    'destination_port_range_max': {
+        'default': None,
+        'convert_to': ext_fc.normalize_port_value
+    }
+}
+
+
 class PortChainNotFound(neutron_exc.NotFound):
     message = _("Port Chain %(id)s not found.")
 
@@ -104,31 +133,21 @@ class PortChainChainIdInConflict(neutron_exc.InvalidInput):
                 "Chain id in port chain %(pc_id)s.")
 
 
-class InvalidChainParameter(neutron_exc.InvalidInput):
-    message = _(
-        "Invalid chain parameter: %%(error_message)s. "
-        "Supported chain parameters are %(supported_paramters)s."
-    ) % {'supported_paramters': SUPPORTED_CHAIN_PARAMETERS}
-
-
-class InvalidServiceFunctionParameter(neutron_exc.InvalidInput):
-    message = _(
-        "Invalid Service function parameter: %%(error_message)s. "
-        "Supported service function parameters are %(supported_paramters)s."
-    ) % {'supported_paramters': SUPPORTED_SF_PARAMETERS}
-
-
-class InvalidPortPairGroupParameter(neutron_exc.InvalidInput):
-    message = _(
-        "Invalid port pair group parameter: %%(error_message)s. "
-        "Supported port pair group parameters are %(supported_paramters)s."
-    ) % {'supported_paramters': SUPPORTED_PPG_PARAMETERS}
+class PortChainInconsistentCorrelations(neutron_exc.InvalidInput):
+    message = _("Port Chain attempted creation included a Port Pair Group "
+                "(%(ppg)s) with a different protocol used as correlation "
+                "type.")
 
 
 class PortPairGroupNotSpecified(neutron_exc.InvalidInput):
     message = _("Port Pair Group is not specified in Port Chain.")
 
 
+class InconsistentCorrelations(neutron_exc.InvalidInput):
+    message = _("Port Pair Group attempted creation included Port Pairs "
+                "with inconsistent correlation types.")
+
+
 class InvalidPortPairGroups(neutron_exc.InUse):
     message = _("Port Pair Group(s) %(port_pair_groups)s in use by "
                 "Port Chain %(port_chain)s.")
@@ -139,7 +158,7 @@ class PortPairPortNotFound(neutron_exc.NotFound):
 
 
 class PortPairIngressEgressDifferentHost(neutron_exc.InvalidInput):
-    message = _("Port Pair ingress port %(ingress)s and"
+    message = _("Port Pair ingress port %(ingress)s and "
                 "egress port %(egress)s not in the same host.")
 
 
@@ -175,10 +194,12 @@ class PortPairInUse(neutron_exc.InUse):
     message = _("Port Pair %(id)s in use.")
 
 
-def normalize_string(value):
-    if value is None:
-        return ''
-    return value
+class PPGParametersInvalidNTupleMappingParameter(neutron_exc.InvalidInput):
+    message = _(
+        "Invalid Port Pair Group N-Tuple Mapping parameters: "
+        "%%(error_message)s. Supported PPG classifier N-Tuple Mapping "
+        "parameters are %(supported_parameters)s."
+    ) % {'supported_parameters': SUPPORTED_PPG_TUPLE_MAPPING}
 
 
 def normalize_port_pair_groups(port_pair_groups):
@@ -189,50 +210,32 @@ def normalize_port_pair_groups(port_pair_groups):
 
 
 def normalize_chain_parameters(parameters):
-    parameters = lib_converters.convert_none_to_empty_dict(parameters)
-    for key in parameters:
-        if key not in SUPPORTED_CHAIN_PARAMETERS:
-            raise InvalidChainParameter(
-                error_message='Unknown key %s.' % key)
-    try:
-        attr.fill_default_value(
-            SUPPORTED_CHAIN_PARAMETERS, parameters)
-        attr.convert_value(
-            SUPPORTED_CHAIN_PARAMETERS, parameters)
-    except ValueError as error:
-        raise InvalidChainParameter(error_message=str(error))
+    if not parameters:
+        return DEFAULT_CHAIN_PARAMETERS
+    if 'correlation' not in parameters:
+        parameters['correlation'] = DEFAULT_CHAIN_CORRELATION
+    if 'symmetric' not in parameters:
+        parameters['symmetric'] = DEFAULT_CHAIN_SYMMETRY
     return parameters
 
 
 def normalize_sf_parameters(parameters):
-    parameters = lib_converters.convert_none_to_empty_dict(parameters)
-    for key in parameters:
-        if key not in SUPPORTED_SF_PARAMETERS:
-            raise InvalidServiceFunctionParameter(
-                error_message='Unknown key %s.' % key)
-    try:
-        attr.fill_default_value(
-            SUPPORTED_SF_PARAMETERS, parameters)
-        attr.convert_value(
-            SUPPORTED_SF_PARAMETERS, parameters)
-    except ValueError as error:
-        raise InvalidServiceFunctionParameter(error_message=str(error))
-    return parameters
+    return parameters if parameters else DEFAULT_SF_PARAMETERS
 
 
 def normalize_ppg_parameters(parameters):
-    parameters = lib_converters.convert_none_to_empty_dict(parameters)
-    for key in parameters:
-        if key not in SUPPORTED_PPG_PARAMETERS:
-            raise InvalidPortPairGroupParameter(
-                error_message='Unknown key %s.' % key)
-    try:
-        attr.fill_default_value(
-            SUPPORTED_PPG_PARAMETERS, parameters)
-        attr.convert_value(
-            SUPPORTED_PPG_PARAMETERS, parameters)
-    except ValueError as error:
-        raise InvalidPortPairGroupParameter(error_message=str(error))
+    if not parameters:
+        return DEFAULT_PPG_PARAMETERS
+    if 'lb_fields' not in parameters:
+        parameters['lb_fields'] = DEFAULT_PPG_LB_FIELDS
+    if 'ppg_n_tuple_mapping' not in parameters:
+        parameters['ppg_n_tuple_mapping'] = DEFAULT_PPG_N_TUPLE
+    if 'ppg_n_tuple_mapping' in parameters:
+        for key, value in parameters['ppg_n_tuple_mapping'].items():
+            for n_key in value:
+                if n_key not in SUPPORTED_PPG_TUPLE_MAPPING:
+                    raise PPGParametersInvalidNTupleMappingParameter(
+                        error_message='Unknow key %s.' % n_key)
     return parameters
 
 
@@ -242,77 +245,111 @@ RESOURCE_ATTRIBUTE_MAP = {
             'allow_post': False, 'allow_put': False,
             'is_visible': True,
             'validate': {'type:uuid': None},
-            'primary_key': True},
+            'primary_key': True
+        },
         'name': {
             'allow_post': True, 'allow_put': True,
-            'is_visible': True, 'default': None,
-            'validate': {'type:string': attr.NAME_MAX_LEN},
-            'convert_to': normalize_string},
+            'is_visible': True, 'default': '',
+            'validate': {'type:string': db_const.NAME_FIELD_SIZE},
+        },
         'description': {
             'allow_post': True, 'allow_put': True,
-            'is_visible': True, 'default': None,
-            'validate': {'type:string': attr.DESCRIPTION_MAX_LEN},
-            'convert_to': normalize_string},
+            'is_visible': True, 'default': '',
+            'validate': {'type:string': db_const.DESCRIPTION_FIELD_SIZE},
+        },
         'tenant_id': {
             'allow_post': True, 'allow_put': False,
             'is_visible': True,
-            'validate': {'type:string': attr.TENANT_ID_MAX_LEN},
-            'required_by_policy': True},
+            'validate': {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
+            'required_by_policy': True
+        },
         'ingress': {
             'allow_post': True, 'allow_put': False,
             'is_visible': True,
-            'validate': {'type:uuid': None}},
+            'validate': {'type:uuid': None}
+        },
         'egress': {
             'allow_post': True, 'allow_put': False,
             'is_visible': True,
-            'validate': {'type:uuid': None}},
+            'validate': {'type:uuid': None}
+        },
         'service_function_parameters': {
             'allow_post': True, 'allow_put': False,
             'is_visible': True, 'default': None,
-            'validate': {'type:dict': None},
-            'convert_to': normalize_sf_parameters},
+            'validate': {
+                'type:dict': {
+                    'correlation': {
+                        'default': DEFAULT_SF_PARAMETERS['correlation'],
+                        'type:values': [None, 'mpls', 'nsh']
+                    },
+                    'weight': {
+                        'default': DEFAULT_SF_PARAMETERS['weight'],
+                        'type:non_negative': None,
+                        'convert_to': lib_converters.convert_to_int
+                    }
+                }
+            },
+            'convert_to': normalize_sf_parameters
+        }
     },
     'port_chains': {
         'id': {
             'allow_post': False, 'allow_put': False,
             'is_visible': True,
             'validate': {'type:uuid': None},
-            'primary_key': True},
+            'primary_key': True
+        },
         'chain_id': {
             'allow_post': True, 'allow_put': False,
             'is_visible': True, 'default': 0,
             'validate': {'type:range': (0, MAX_CHAIN_ID)},
-            'convert_to': lib_converters.convert_to_int},
+            'convert_to': lib_converters.convert_to_int
+        },
         'name': {
             'allow_post': True, 'allow_put': True,
-            'is_visible': True, 'default': None,
-            'validate': {'type:string': attr.NAME_MAX_LEN},
-            'convert_to': normalize_string},
+            'is_visible': True, 'default': '',
+            'validate': {'type:string': db_const.NAME_FIELD_SIZE},
+        },
         'description': {
             'allow_post': True, 'allow_put': True,
-            'is_visible': True, 'default': None,
-            'validate': {'type:string': attr.DESCRIPTION_MAX_LEN},
-            'convert_to': normalize_string},
+            'is_visible': True, 'default': '',
+            'validate': {'type:string': db_const.DESCRIPTION_FIELD_SIZE},
+        },
         'tenant_id': {
             'allow_post': True, 'allow_put': False,
             'is_visible': True,
-            'validate': {'type:string': attr.TENANT_ID_MAX_LEN},
-            'required_by_policy': True},
+            'validate': {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
+            'required_by_policy': True
+        },
         'port_pair_groups': {
             'allow_post': True, 'allow_put': True,
             'is_visible': True,
             'validate': {'type:uuid_list': None},
-            'convert_to': normalize_port_pair_groups},
+            'convert_to': normalize_port_pair_groups
+        },
         'flow_classifiers': {
             'allow_post': True, 'allow_put': True,
             'is_visible': True, 'default': None,
             'validate': {'type:uuid_list': None},
-            'convert_to': lib_converters.convert_to_list},
+            'convert_to': lib_converters.convert_to_list
+        },
         'chain_parameters': {
             'allow_post': True, 'allow_put': False,
             'is_visible': True, 'default': None,
-            'validate': {'type:dict': None},
-            'convert_to': normalize_chain_parameters}
+            'validate': {
+                'type:dict': {
+                    'correlation': {
+                        'default': DEFAULT_CHAIN_PARAMETERS['correlation'],
+                        'type:values': ['mpls', 'nsh']
+                    },
+                    'symmetric': {
+                        'default': DEFAULT_CHAIN_PARAMETERS['symmetric'],
+                        'convert_to': lib_converters.convert_to_boolean
+                    }
+                }
+            },
+            'convert_to': normalize_chain_parameters
+        }
     },
     'port_pair_groups': {
         'id': {
@@ -322,33 +359,65 @@ RESOURCE_ATTRIBUTE_MAP = {
             'primary_key': True},
         'group_id': {
             'allow_post': False, 'allow_put': False,
-            'is_visible': True},
+            'is_visible': True
+        },
         'name': {
             'allow_post': True, 'allow_put': True,
-            'is_visible': True, 'default': None,
-            'validate': {'type:string': attr.NAME_MAX_LEN},
-            'convert_to': normalize_string},
+            'is_visible': True, 'default': '',
+            'validate': {'type:string': db_const.NAME_FIELD_SIZE},
+        },
         'description': {
             'allow_post': True, 'allow_put': True,
-            'is_visible': True, 'default': None,
-            'validate': {'type:string': attr.DESCRIPTION_MAX_LEN},
-            'convert_to': normalize_string},
+            'is_visible': True, 'default': '',
+            'validate': {'type:string': db_const.DESCRIPTION_FIELD_SIZE},
+        },
         'tenant_id': {
             'allow_post': True, 'allow_put': False,
             'is_visible': True,
-            'validate': {'type:string': attr.TENANT_ID_MAX_LEN},
-            'required_by_policy': True},
+            'validate': {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
+            'required_by_policy': True
+        },
         'port_pairs': {
             'allow_post': True, 'allow_put': True,
             'is_visible': True, 'default': None,
             'validate': {'type:uuid_list': None},
-            'convert_to': lib_converters.convert_none_to_empty_list},
+            'convert_to': lib_converters.convert_none_to_empty_list
+        },
         'port_pair_group_parameters': {
             'allow_post': True, 'allow_put': False,
             'is_visible': True, 'default': None,
-            'validate': {'type:dict': None},
-            'convert_to': normalize_ppg_parameters},
-    },
+            'validate': {
+                'type:dict': {
+                    'lb_fields': {
+                        'default': DEFAULT_PPG_PARAMETERS['lb_fields'],
+                        'type:list_of_allowed_values': SUPPORTED_LB_FIELDS
+                    },
+                    'ppg_n_tuple_mapping': {
+                        'default': DEFAULT_PPG_PARAMETERS[
+                            'ppg_n_tuple_mapping'],
+                        'validate': {
+                            'type:dict': {
+                                'ingress_n_tuple': {
+                                    'default': {},
+                                    'validate': {
+                                        'type:dict': ppg_n_tuple_validact_dict
+                                    }
+                                },
+                                'egress_n_tuple': {
+                                    'default': {},
+                                    'validate': {
+                                        'type:dict': ppg_n_tuple_validact_dict
+                                    }
+                                }
+                            }
+                        },
+                        'convert_to': lib_converters.convert_none_to_empty_dict
+                    }
+                }
+            },
+            'convert_to': normalize_ppg_parameters
+        }
+    }
 }
 
 sfc_quota_opts = [
@@ -369,7 +438,7 @@ sfc_quota_opts = [
 cfg.CONF.register_opts(sfc_quota_opts, 'QUOTAS')
 
 
-class Sfc(neutron_ext.ExtensionDescriptor):
+class Sfc(extensions.ExtensionDescriptor):
     """Service Function Chain extension."""
 
     @classmethod
@@ -402,7 +471,6 @@ class Sfc(neutron_ext.ExtensionDescriptor):
         plural_mappings = resource_helper.build_plural_mappings(
             {}, RESOURCE_ATTRIBUTE_MAP)
         plural_mappings['sfcs'] = 'sfc'
-        attr.PLURALS.update(plural_mappings)
         return resource_helper.build_resource_info(
             plural_mappings,
             RESOURCE_ATTRIBUTE_MAP,
@@ -412,8 +480,7 @@ class Sfc(neutron_ext.ExtensionDescriptor):
     def get_extended_resources(self, version):
         if version == "2.0":
             return RESOURCE_ATTRIBUTE_MAP
-        else:
-            return {}
+        return {}
 
 
 @six.add_metaclass(ABCMeta)
