@@ -16,9 +16,10 @@ from oslo_log import helpers as log_helpers
 from oslo_log import log as logging
 from oslo_utils import excutils
 
-from neutron import manager
+from neutron_lib.plugins import directory
+
+from neutron.db import api as db_api
 
-from networking_sfc._i18n import _LE
 from networking_sfc.db import flowclassifier_db as fc_db
 from networking_sfc.extensions import flowclassifier as fc_ext
 from networking_sfc.services.flowclassifier.common import context as fc_ctx
@@ -42,12 +43,11 @@ class FlowClassifierPlugin(fc_db.FlowClassifierDbPlugin):
 
     def _get_port(self, context, id):
         port = super(FlowClassifierPlugin, self)._get_port(context, id)
-        core_plugin = manager.NeutronManager.get_plugin()
-        return core_plugin.get_port(context, port['id'])
+        return directory.get_plugin().get_port(context, port['id'])
 
     @log_helpers.log_method_call
     def create_flow_classifier(self, context, flow_classifier):
-        with context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(context):
             fc_db = super(FlowClassifierPlugin, self).create_flow_classifier(
                 context, flow_classifier)
             fc_db_context = fc_ctx.FlowClassifierContext(self, context, fc_db)
@@ -55,33 +55,37 @@ class FlowClassifierPlugin(fc_db.FlowClassifierDbPlugin):
                 fc_db_context)
 
         try:
-            self.driver_manager.create_flow_classifier(fc_db_context)
+            self.driver_manager.create_flow_classifier_postcommit(
+                fc_db_context)
         except fc_exc.FlowClassifierDriverError as e:
             LOG.exception(e)
             with excutils.save_and_reraise_exception():
-                LOG.error(_LE("Create flow classifier failed, "
-                              "deleting flow_classifier '%s'"),
+                LOG.error("Create flow classifier failed, "
+                          "deleting flow_classifier '%s'",
                           fc_db['id'])
                 self.delete_flow_classifier(context, fc_db['id'])
         return fc_db
 
     @log_helpers.log_method_call
     def update_flow_classifier(self, context, id, flow_classifier):
-        original_flowclassifier = self.get_flow_classifier(context, id)
-        updated_fc = super(
-            FlowClassifierPlugin, self
-        ).update_flow_classifier(
-            context, id, flow_classifier)
-        fc_db_context = fc_ctx.FlowClassifierContext(
-            self, context, updated_fc,
-            original_flowclassifier=original_flowclassifier)
+        with db_api.context_manager.writer.using(context):
+            original_flowclassifier = self.get_flow_classifier(context, id)
+            updated_fc = super(
+                FlowClassifierPlugin, self
+            ).update_flow_classifier(
+                context, id, flow_classifier)
+            fc_db_context = fc_ctx.FlowClassifierContext(
+                self, context, updated_fc,
+                original_flowclassifier=original_flowclassifier)
+            self.driver_manager.update_flow_classifier_precommit(fc_db_context)
         try:
-            self.driver_manager.update_flow_classifier(fc_db_context)
+            self.driver_manager.update_flow_classifier_postcommit(
+                fc_db_context)
         except fc_exc.FlowClassifierDriverError as e:
             LOG.exception(e)
             with excutils.save_and_reraise_exception():
-                LOG.error(_LE("Update flow classifier failed, "
-                              "flow_classifier '%s'"),
+                LOG.error("Update flow classifier failed, "
+                          "flow_classifier '%s'",
                           updated_fc['id'])
 
         return updated_fc
@@ -95,9 +99,14 @@ class FlowClassifierPlugin(fc_db.FlowClassifierDbPlugin):
         except fc_exc.FlowClassfierDriverError as e:
             LOG.exception(e)
             with excutils.save_and_reraise_exception():
-                LOG.error(_LE("Delete flow classifier failed, "
-                              "flow_classifier '%s'"),
+                LOG.error("Delete flow classifier failed, "
+                          "flow_classifier '%s'",
                           fc_id)
 
-        super(FlowClassifierPlugin, self).delete_flow_classifier(
-            context, fc_id)
+        with db_api.context_manager.writer.using(context):
+            fc = self.get_flow_classifier(context, fc_id)
+            fc_context = fc_ctx.FlowClassifierContext(self, context, fc)
+            super(FlowClassifierPlugin, self).delete_flow_classifier(
+                context, fc_id)
+            self.driver_manager.delete_flow_classifier_precommit(fc_context)
+        self.driver_manager.delete_flow_classifier_postcommit(fc_context)
