@@ -52,6 +52,14 @@ PC_INGRESS_PRI = 30
 # Reverse group number offset for dump_group
 REVERSE_GROUP_NUMBER_OFFSET = 7000
 
+TAP_CLASSIFIER_TABLE = 7
+# This table floods TAP packets on tunnel ports
+TAP_TUNNEL_OUTPUT_TABLE = 25
+
+# actions
+RESUBMIT_TAP_TABLE = ',resubmit(,%s)' % TAP_CLASSIFIER_TABLE
+NORMAL_ACTION = ",NORMAL"
+
 
 class SfcOVSAgentDriver(sfc.SfcAgentDriver):
     """This class will support MPLS frame
@@ -63,17 +71,11 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
     +-------------------------------+---------------+--------------------+
     """
 
-    REQUIRED_PROTOCOLS = [
-        ovs_consts.OPENFLOW10,
-        ovs_consts.OPENFLOW11,
-        ovs_consts.OPENFLOW12,
-        ovs_consts.OPENFLOW13,
-    ]
-
     def __init__(self):
         super(SfcOVSAgentDriver, self).__init__()
         self.agent_api = None
         self.br_int = None
+        self.br_tun = None
 
         self.local_ip = None
         self.patch_tun_ofport = None
@@ -85,7 +87,6 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
     def initialize(self):
         self.br_int = ovs_ext_lib.SfcOVSBridgeExt(
             self.agent_api.request_int_br())
-        self.br_int.set_protocols(SfcOVSAgentDriver.REQUIRED_PROTOCOLS)
 
         self.local_ip = cfg.CONF.OVS.local_ip
         self.patch_tun_ofport = self.br_int.get_port_ofport(
@@ -94,6 +95,12 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
 
         self._clear_sfc_flow_on_int_br()
 
+        self.br_tun = ovs_ext_lib.SfcOVSBridgeExt(
+            self.agent_api.request_tun_br())
+        self.patch_int_ofport = self.br_tun.get_port_ofport(
+            cfg.CONF.OVS.tun_peer_patch_port)
+        self._clear_sfc_flow_on_tun_br()
+
     def update_flow_rules(self, flowrule, flowrule_status):
         if flowrule['fwd_path'] is False and flowrule['node_type'] == \
                 'sf_node':
@@ -104,7 +111,8 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
 
             if flowrule.get('egress', None):
                 self._setup_egress_flow_rules(flowrule)
-            if flowrule.get('ingress', None):
+            if flowrule.get('ingress', None) and not flowrule.get(
+                    'skip_ingress_flow_config', None):
                 self._setup_ingress_flow_rules(flowrule)
             flowrule_status_temp = {'id': flowrule['id'],
                                     'status': constants.STATUS_ACTIVE}
@@ -140,15 +148,7 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
                     else:
                         self.br_int.delete_group(group_id=group_id +
                                                  REVERSE_GROUP_NUMBER_OFFSET)
-                    for item in flowrule['next_hops']:
-                        if flowrule['fwd_path']:
-                            self.br_int.delete_flows(
-                                table=ACROSS_SUBNET_TABLE,
-                                dl_dst=item['in_mac_address'])
-                        else:
-                            self.br_int.delete_flows(
-                                table=ACROSS_SUBNET_TABLE,
-                                dl_dst=item['mac_address'])
+                    self._delete_across_subnet_table_flows(flowrule)
 
             if flowrule['ingress'] is not None:
                 # delete table INGRESS_TABLE ingress match flow rule
@@ -159,6 +159,8 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
                     # for ingress traffic
                     if pc_corr == 'mpls':
                         self._delete_flows_mpls(flowrule, vif_port)
+                    elif pc_corr == 'nsh':
+                        self._delete_flows_nsh(flowrule, vif_port)
         except Exception as e:
             flowrule_status_temp = {'id': flowrule['id'],
                                     'status': constants.STATUS_ERROR}
@@ -170,13 +172,17 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
         self.br_int.delete_group(group_id='all')
         self.br_int.delete_flows(table=ACROSS_SUBNET_TABLE)
         self.br_int.delete_flows(table=INGRESS_TABLE)
+        self.br_int.delete_flows(table=TAP_CLASSIFIER_TABLE)
         self.br_int.install_goto(dest_table_id=INGRESS_TABLE,
                                  priority=PC_DEF_PRI,
                                  eth_type=constants.ETH_TYPE_MPLS)
+        self.br_int.install_goto(dest_table_id=INGRESS_TABLE,
+                                 priority=PC_DEF_PRI,
+                                 eth_type=constants.ETH_TYPE_NSH)
         self.br_int.install_drop(table_id=INGRESS_TABLE)
 
     def _parse_flow_classifier(self, flow_classifier):
-        dl_type, nw_proto, source_port_masks, destination_port_masks = (
+        eth_type, nw_proto, source_port_masks, destination_port_masks = (
             (None, ) * 4)
 
         if (not flow_classifier['source_port_range_min'] and
@@ -214,7 +220,7 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
                 flow_classifier['destination_port_range_max'])
 
         if flow_classifier['ethertype'] == "IPv4":
-            dl_type = constants.ETH_TYPE_IP
+            eth_type = constants.ETH_TYPE_IP
             if n_consts.PROTO_NAME_TCP == flow_classifier['protocol']:
                 nw_proto = n_consts.PROTO_NUM_TCP
             elif n_consts.PROTO_NAME_UDP == flow_classifier['protocol']:
@@ -227,7 +233,8 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
             LOG.error("Current portchain agent doesn't support IPv6")
         else:
             LOG.error("invalid protocol input")
-        return (dl_type, nw_proto,
+
+        return (eth_type, nw_proto,
                 source_port_masks,
                 destination_port_masks)
 
@@ -240,7 +247,7 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
             return flow_infos
 
         # parse and transfer flow info to match field info
-        dl_type, nw_proto, source_port_masks, destination_port_masks = (
+        eth_type, nw_proto, source_port_masks, destination_port_masks = (
             self._parse_flow_classifier(flow_classifier))
 
         if flowrule['fwd_path']:
@@ -271,19 +278,14 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
                     else:
                         tp_dst = '%s' % source_port
                         tp_src = '%s' % destination_port
-                    if nw_proto is None:
-                        flow_infos.append({'dl_type': dl_type,
-                                           'nw_src': nw_src,
-                                           'nw_dst': nw_dst,
-                                           'tp_src': tp_src,
-                                           'tp_dst': tp_dst})
-                    else:
-                        flow_infos.append({'dl_type': dl_type,
-                                           'nw_proto': nw_proto,
-                                           'nw_src': nw_src,
-                                           'nw_dst': nw_dst,
-                                           'tp_src': tp_src,
-                                           'tp_dst': tp_dst})
+                    flow_info = {'eth_type': eth_type,
+                                 'nw_src': nw_src,
+                                 'nw_dst': nw_dst,
+                                 'tp_src': tp_src,
+                                 'tp_dst': tp_dst}
+                    if nw_proto:
+                        flow_info['nw_proto'] = nw_proto
+                    flow_infos.append(flow_info)
 
         return flow_infos
 
@@ -300,15 +302,27 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
 
         return flow_infos
 
+    def _match_by_header(self, match_info, nsp, nsi):
+        match_info['reg0'] = (nsp << 8) | nsi
+        # on header-matching there's no in_port
+        match_info.pop('in_port', None)
+
     def _setup_local_switch_flows_on_int_br(self, flowrule,
                                             flow_classifier_list, actions,
                                             add_flow=True, match_inport=True):
         inport_match = {}
         priority = PC_DEF_PRI
-
         # no pp_corr means that classification will not be based on encap
-        pp_corr = flowrule.get('pp_corr', None)
+        pp_corr = flowrule.get('pp_corr')
         node_type = flowrule['node_type']
+        branch_info = flowrule.get('branch_info')
+        on_add = None
+        flow_count = 1
+        if branch_info and node_type == constants.SRC_NODE:
+            # for branching, we need as many flows (per flow info) as branches
+            # because we can't AND-match the same field in a single flow
+            flow_count = len(branch_info.get('matches'))
+            on_add = branch_info.get('on_add')
 
         if match_inport is True:
             egress_port = self.br_int.get_vif_port_by_id(flowrule['egress'])
@@ -326,17 +340,29 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
                     if pp_corr == 'mpls':
                         match_info = self._build_classification_match_sfc_mpls(
                             flowrule, match_info)
+                    elif pp_corr == 'nsh':
+                        match_info = self._build_classification_match_sfc_nsh(
+                            flowrule, match_info)
 
-            if add_flow:
-                self.br_int.add_flow(table=ovs_consts.LOCAL_SWITCHING,
-                                     priority=priority,
-                                     actions=actions,
-                                     **match_info)
-            else:
-                self.br_int.delete_flows(table=ovs_consts.LOCAL_SWITCHING,
+            for i in range(flow_count):
+                if branch_info:
+                    # for Service Graphs (branching):
+                    nsp = branch_info['matches'][i][0]
+                    nsi = branch_info['matches'][i][1]
+                if add_flow:
+                    if on_add:
+                        self._match_by_header(match_info, nsp, nsi)
+                    self.br_int.add_flow(table=ovs_consts.LOCAL_SWITCHING,
                                          priority=priority,
-                                         strict=True,
+                                         actions=actions,
                                          **match_info)
+                else:
+                    if on_add is False:
+                        self._match_by_header(match_info, nsp, nsi)
+                    self.br_int.delete_flows(table=ovs_consts.LOCAL_SWITCHING,
+                                             priority=priority,
+                                             strict=True,
+                                             **match_info)
 
     def _setup_egress_flow_rules(self, flowrule, match_inport=True):
         group_id = flowrule.get('next_group_id', None)
@@ -344,6 +370,7 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
         pc_corr = flowrule.get('pc_corr', 'mpls')
         pp_corr = flowrule.get('pp_corr', None)
         node_type = flowrule.get('node_type')
+        next_hop_tap_enabled = None
         # if the group is not none, install the egress rule for this SF
         if group_id and next_hops:
             # 1st, install br-int flow rule on table ACROSS_SUBNET_TABLE
@@ -351,6 +378,9 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
             buckets = []
             vlan = self._get_vlan_by_port(flowrule['egress'])
 
+            if isinstance(next(iter(next_hops)), dict):
+                next_hop_tap_enabled = next_hops[0].get('tap_enabled')
+
             for item in next_hops:
                 # all next hops share same pp_corr, enforced by higher layers
                 pp_corr_nh = item.get('pp_corr', None)
@@ -377,7 +407,10 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
                     if pc_corr == 'mpls':
                         push_encap = self._build_push_mpls(flowrule['nsp'],
                                                            flowrule['nsi'])
-                        across_flow = push_encap + across_flow
+                    elif pc_corr == 'nsh':
+                        push_encap = self._build_push_nsh(flowrule['nsp'],
+                                                          flowrule['nsi'])
+                    across_flow = push_encap + across_flow
 
                 subnet_actions_list.append(across_flow)
 
@@ -388,47 +421,24 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
                     subnet_actions = 'output:%s' % self.patch_tun_ofport
                 subnet_actions_list.append(subnet_actions)
 
-                dl_type = constants.ETH_TYPE_IP
+                eth_type = constants.ETH_TYPE_IP
                 if pp_corr == 'mpls' or pp_corr_nh == 'mpls':
-                    dl_type = constants.ETH_TYPE_MPLS
+                    eth_type = constants.ETH_TYPE_MPLS
+                elif pp_corr == 'nsh' or pp_corr_nh == 'nsh':
+                    eth_type = constants.ETH_TYPE_NSH
 
-                if flowrule['fwd_path']:
-                    self.br_int.add_flow(
-                        table=ACROSS_SUBNET_TABLE,
-                        priority=0,
-                        dl_dst=item['in_mac_address'],
-                        dl_type=dl_type,
-                        actions="%s" % ','.join(subnet_actions_list))
+                if item.get('tap_enabled'):
+                    self._add_tap_classification_flows(flowrule,
+                                                       item,
+                                                       subnet_actions_list)
                 else:
-                    self.br_int.add_flow(
-                        table=ACROSS_SUBNET_TABLE,
-                        priority=0,
-                        dl_dst=item['mac_address'],
-                        dl_type=dl_type,
-                        actions="%s" % ','.join(subnet_actions_list))
+                    self._configure_across_subnet_flow(flowrule,
+                                                       item,
+                                                       subnet_actions_list,
+                                                       eth_type)
 
-            group_content = self.br_int.dump_group_for_id(group_id)
-            buckets = ','.join(buckets)
-            if flowrule['fwd_path']:
-                if group_content.find('group_id=%d' % group_id) == -1:
-                    self.br_int.add_group(group_id=group_id,
-                                          type='select',
-                                          buckets=buckets)
-                else:
-                    self.br_int.mod_group(group_id=group_id,
-                                          type='select',
-                                          buckets=buckets)
-            else:
-                # set different id for rev_group
-                rev_group_id = group_id + REVERSE_GROUP_NUMBER_OFFSET
-                if group_content.find('group_id=%d' % (rev_group_id)) == -1:
-                    self.br_int.add_group(group_id=rev_group_id,
-                                          type='select',
-                                          buckets=buckets)
-                else:
-                    self.br_int.mod_group(group_id=rev_group_id,
-                                          type='select',
-                                          buckets=buckets)
+            if not next_hop_tap_enabled:
+                self._add_group_table(buckets, flowrule, group_id)
 
             # 2nd, install br-int flow rule on table 0 for egress traffic
             enc_actions = ""
@@ -438,11 +448,17 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
                 if pc_corr == 'mpls':
                     enc_actions = self._build_push_mpls(flowrule['nsp'],
                                                         flowrule['nsi'])
+                elif pc_corr == 'nsh':
+                    enc_actions = self._build_push_nsh(flowrule['nsp'],
+                                                       flowrule['nsi'])
             if flowrule['fwd_path']:
-                enc_actions = enc_actions + ("group:%d" % group_id)
+                enc_actions += "group:%d" % group_id
             else:
-                enc_actions = enc_actions + ("group:%d" % rev_group_id)
+                rev_group_id = group_id + REVERSE_GROUP_NUMBER_OFFSET
+                enc_actions += "group:%d" % rev_group_id
 
+            enc_actions = self._update_enc_actions(enc_actions, flowrule,
+                                                   next_hop_tap_enabled)
             # to uninstall the removed flow classifiers
             self._setup_local_switch_flows_on_int_br(
                 flowrule,
@@ -461,25 +477,50 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
             end_of_chain_actions = 'normal'
             # at the end of the chain, the header must be removed (if used)
             if (node_type != constants.SRC_NODE) and pp_corr:
-                if pc_corr == 'mpls':
-                    end_of_chain_actions = ("pop_mpls:0x%04x,%s" % (
-                                            constants.ETH_TYPE_IP,
-                                            end_of_chain_actions))
-            # to uninstall the removed flow classifiers
-            self._setup_local_switch_flows_on_int_br(
-                flowrule,
-                flowrule['del_fcs'],
-                None,
-                add_flow=False,
-                match_inport=True)
+                branch_point = flowrule.get('branch_point')
+                if branch_point:
+                    nsp = flowrule['nsp']
+                    nsi = flowrule['nsi']
+                    sfpi = (nsp << 8) | nsi
+                    if pc_corr == 'mpls':
+                        end_of_chain_actions = (
+                            'load:%s->NXM_NX_REG0[],'
+                            'pop_mpls:0x%04x,resubmit(,0)' % (
+                                hex(sfpi), constants.ETH_TYPE_IP))
+                    elif pc_corr == 'nsh':
+                        end_of_chain_actions = (
+                            "load:%s->NXM_NX_REG0[],"
+                            "decap(),decap(),resubmit(,0)" % (
+                                hex(sfpi)))
+                else:
+                    if pc_corr == 'mpls':
+                        end_of_chain_actions = ("pop_mpls:0x%04x,%s" % (
+                                                constants.ETH_TYPE_IP,
+                                                end_of_chain_actions))
+                    elif pc_corr == 'nsh':
+                        end_of_chain_actions = ("decap(),decap(),%s" % (
+                                                end_of_chain_actions))
 
-            # to install the added flow classifiers
-            self._setup_local_switch_flows_on_int_br(
-                flowrule,
-                flowrule['add_fcs'],
-                actions=end_of_chain_actions,
-                add_flow=True,
-                match_inport=True)
+            if flowrule.get('tap_enabled'):
+                end_of_chain_actions += RESUBMIT_TAP_TABLE
+
+            # to uninstall the removed flow classifiers
+            if 'del_fcs' in flowrule:
+                self._setup_local_switch_flows_on_int_br(
+                    flowrule,
+                    flowrule['del_fcs'],
+                    None,
+                    add_flow=False,
+                    match_inport=True)
+
+            if 'add_fcs' in flowrule:
+                # to install the added flow classifiers
+                self._setup_local_switch_flows_on_int_br(
+                    flowrule,
+                    flowrule['add_fcs'],
+                    actions=end_of_chain_actions,
+                    add_flow=True,
+                    match_inport=True)
 
     def _get_vlan_by_port(self, port_id):
         try:
@@ -496,22 +537,38 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
             pp_corr = flowrule['pp_corr']
 
             # install br-int flow rule on table 0 for ingress traffic
+            # install an SFC Proxy if the port pair doesn't support the
+            # SFC encapsulation (pc_corr) specified in the chain
             if pc_corr == 'mpls':
+                if flowrule.get('tap_enabled'):
+                    return self._add_tap_ingress_flow(flowrule, vif_port, vlan)
                 if pp_corr is None:
-                    # install an SFC Proxy if the port pair doesn't support the
-                    # SFC encapsulation (pc_corr) specified in the chain
                     match_field = self._build_proxy_sfc_mpls(flowrule,
                                                              vif_port, vlan)
                 elif pp_corr == 'mpls':
                     match_field = self._build_forward_sfc_mpls(flowrule,
                                                                vif_port, vlan)
+            elif pc_corr == 'nsh':
+                if pp_corr is None:
+                    match_field = self._build_proxy_sfc_nsh(flowrule,
+                                                            vif_port, vlan)
+                elif pp_corr == 'nsh':
+                    match_field = self._build_forward_sfc_nsh(flowrule,
+                                                              vif_port, vlan)
             self.br_int.add_flow(**match_field)
 
     def _build_classification_match_sfc_mpls(self, flowrule, match_info):
-        match_info['dl_type'] = constants.ETH_TYPE_MPLS
+        match_info['eth_type'] = constants.ETH_TYPE_MPLS
         match_info['mpls_label'] = flowrule['nsp'] << 8 | flowrule['nsi']
         return match_info
 
+    def _build_classification_match_sfc_nsh(self, flowrule, match_info):
+        match_info['eth_type'] = constants.ETH_TYPE_NSH
+        match_info['nsh_mdtype'] = 1
+        match_info['nsh_spi'] = flowrule['nsp']
+        match_info['nsh_si'] = flowrule['nsi']
+        return match_info
+
     def _build_push_mpls(self, nsp, nsi):
         return (
             "push_mpls:0x%04x,"
@@ -519,6 +576,13 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
             "set_mpls_ttl:%d," %
             (constants.ETH_TYPE_MPLS, nsp << 8 | nsi, nsi))
 
+    def _build_push_nsh(self, nsp, nsi):
+        return (
+            "encap(hdr=nsh,prop(class=nsh,type=md_type,val=1)),"
+            "set_field:%s->nsh_spi,set_field:%s->nsh_si,"
+            "encap(hdr=ethernet)," %
+            (hex(nsp), hex(nsi)))
+
     def _build_ingress_common_match_field(self, vif_port, vlan):
         return dict(
             table=INGRESS_TABLE,
@@ -528,10 +592,18 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
 
     def _build_ingress_match_field_sfc_mpls(self, flowrule, vif_port, vlan):
         match_field = self._build_ingress_common_match_field(vif_port, vlan)
-        match_field['dl_type'] = constants.ETH_TYPE_MPLS
+        match_field['eth_type'] = constants.ETH_TYPE_MPLS
         match_field['mpls_label'] = flowrule['nsp'] << 8 | flowrule['nsi'] + 1
         return match_field
 
+    def _build_ingress_match_field_sfc_nsh(self, flowrule, vif_port, vlan):
+        match_field = self._build_ingress_common_match_field(vif_port, vlan)
+        match_field['eth_type'] = constants.ETH_TYPE_NSH
+        match_field['nsh_mdtype'] = 1
+        match_field['nsh_spi'] = flowrule['nsp']
+        match_field['nsh_si'] = flowrule['nsi'] + 1
+        return match_field
+
     def _build_proxy_sfc_mpls(self, flowrule, vif_port, vlan):
         match_field = self._build_ingress_match_field_sfc_mpls(
             flowrule, vif_port, vlan)
@@ -540,6 +612,16 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
         match_field['actions'] = actions
         return match_field
 
+    def _build_proxy_sfc_nsh(self, flowrule, vif_port, vlan):
+        match_field = self._build_ingress_match_field_sfc_nsh(
+            flowrule, vif_port, vlan)
+        actions = ("strip_vlan,move:NXM_OF_ETH_DST->OXM_OF_PKT_REG[0..47],"
+                   "decap(),decap(),"
+                   "move:OXM_OF_PKT_REG[0..47]->NXM_OF_ETH_DST,output:%s"
+                   "" % vif_port.ofport)
+        match_field['actions'] = actions
+        return match_field
+
     def _build_forward_sfc_mpls(self, flowrule, vif_port, vlan):
         match_field = self._build_ingress_match_field_sfc_mpls(
             flowrule, vif_port, vlan)
@@ -547,10 +629,260 @@ class SfcOVSAgentDriver(sfc.SfcAgentDriver):
         match_field['actions'] = actions
         return match_field
 
+    def _build_forward_sfc_nsh(self, flowrule, vif_port, vlan):
+        match_field = self._build_ingress_match_field_sfc_nsh(
+            flowrule, vif_port, vlan)
+        actions = ("strip_vlan, output:%s" % vif_port.ofport)
+        match_field['actions'] = actions
+        return match_field
+
     def _delete_flows_mpls(self, flowrule, vif_port):
+        if flowrule.get('tap_enabled'):
+            self.br_int.delete_flows(
+                table=INGRESS_TABLE,
+                eth_type=constants.ETH_TYPE_MPLS,
+                dl_src=flowrule['mac_address'],
+                mpls_label=flowrule['nsp'] << 8 | flowrule['nsi']
+            )
+        else:
+            self.br_int.delete_flows(
+                table=INGRESS_TABLE,
+                eth_type=constants.ETH_TYPE_MPLS,
+                dl_dst=vif_port.vif_mac,
+                mpls_label=flowrule['nsp'] << 8 | flowrule['nsi'] + 1
+            )
+
+    def _add_group_table(self, buckets, flowrule, group_id):
+        group_content = self.br_int.dump_group_for_id(group_id)
+        buckets = ','.join(buckets)
+        if flowrule['fwd_path']:
+            if group_content.find('group_id=%d' % group_id) == -1:
+                self.br_int.add_group(group_id=group_id,
+                                      type='select',
+                                      buckets=buckets)
+            else:
+                self.br_int.mod_group(group_id=group_id,
+                                      type='select',
+                                      buckets=buckets)
+        else:
+            # set different id for rev_group
+            rev_group_id = group_id + REVERSE_GROUP_NUMBER_OFFSET
+            if group_content.find('group_id=%d' % (rev_group_id)) == -1:
+                self.br_int.add_group(group_id=rev_group_id,
+                                      type='select',
+                                      buckets=buckets)
+            else:
+                self.br_int.mod_group(group_id=rev_group_id,
+                                      type='select',
+                                      buckets=buckets)
+
+    def _configure_across_subnet_flow(self, flowrule, item,
+                                      subnet_actions_list, eth_type):
+        if flowrule['fwd_path']:
+            self.br_int.add_flow(
+                table=ACROSS_SUBNET_TABLE,
+                priority=0,
+                dl_dst=item['in_mac_address'],
+                eth_type=eth_type,
+                actions="%s" % ','.join(subnet_actions_list))
+        else:
+            self.br_int.add_flow(
+                table=ACROSS_SUBNET_TABLE,
+                priority=0,
+                dl_dst=item['mac_address'],
+                eth_type=eth_type,
+                actions="%s" % ','.join(subnet_actions_list))
+
+    def _add_tap_classification_flows(self, flowrule, item,
+                                      subnet_actions_list):
+        egress_port = self.br_int.get_vif_port_by_id(flowrule['egress'])
+        vlan = self._get_vlan_by_port(flowrule['egress'])
+        if not egress_port:
+            return
+        in_port = egress_port.ofport
+        vif_mac = egress_port.vif_mac
+
+        tap_action = ""
+        if flowrule['pc_corr'] == 'mpls':
+            tap_action += self._build_push_mpls(item['nsp'], item['nsi'])
+        tap_action += "mod_vlan_vid:%d," % vlan
+        subnet_actions_list[0] = tap_action
+        ovs_rule = dict()
+
+        self._get_eth_type(flowrule, item, ovs_rule)
+        ovs_rule.update(table=TAP_CLASSIFIER_TABLE,
+                        priority=0,
+                        in_port=in_port,
+                        dl_src=vif_mac,
+                        actions="%s" % ''.join(subnet_actions_list)
+                        )
+        self.br_int.add_flow(**ovs_rule)
+        if item['local_endpoint'] != self.local_ip:
+            self._configure_tunnel_bridge_flows(flowrule, item, vif_mac)
+
+    def _get_eth_type(self, flowrule, item, ovs_rule):
+        # eth_type is decided based on current node's pp_corr and next node of
+        # Tap node's pp_corr
+        if flowrule['pp_corr'] == item.get('pp_corr_tap_nh') is None:
+            ovs_rule.update(eth_type=constants.ETH_TYPE_IP)
+        elif flowrule['pp_corr'] and not item.get('pp_corr_tap_nh'):
+            ovs_rule.update(eth_type=constants.ETH_TYPE_IP)
+        elif not flowrule['pp_corr'] and item.get('pp_corr_tap_nh'):
+            if item['pp_corr_tap_nh'] == 'mpls':
+                eth_type = constants.ETH_TYPE_MPLS
+                mpls_label = flowrule['nsp'] << 8 | flowrule['nsi']
+                ovs_rule.update(mpls_label=mpls_label,
+                                eth_type=eth_type)
+        else:
+            if flowrule['pp_corr'] == 'mpls' or item.get(
+                    'pp_corr_tap_nh') == 'mpls':
+                ovs_rule.update(eth_type=constants.ETH_TYPE_MPLS)
+
+    def _configure_tunnel_bridge_flows(self, flowrule, item, vif_mac):
+        local_tunnel_ports = [port for port in
+                              self.br_tun.get_bridge_ports()
+                              if port != self.patch_int_ofport]
+        match_info = {'in_port': self.patch_int_ofport,
+                      'dl_src': vif_mac}
+        if flowrule['pc_corr'] == 'mpls':
+            self._build_classification_match_sfc_mpls(item, match_info)
+        self.br_tun.add_flow(
+            table=0,
+            priority=30,
+            actions="resubmit(,%s)" % TAP_TUNNEL_OUTPUT_TABLE,
+            **match_info
+        )
+        output_actions = "strip_vlan,load:%s->NXM_NX_TUN_ID[]" % (
+            hex(flowrule['segment_id']))
+        for port in local_tunnel_ports:
+            output_actions += (",output:%d" % port)
+        self.br_tun.add_flow(
+            table=TAP_TUNNEL_OUTPUT_TABLE,
+            priority=0,
+            actions=output_actions,
+            **match_info
+        )
+
+    def _build_buckets(self, buckets, flowrule, item):
+        if item.get('tap_enabled'):
+            # Tap PPG doesn't use bucket as of now.
+            return
+        if flowrule['fwd_path']:
+            bucket = (
+                'bucket=weight=%d, mod_dl_dst:%s, resubmit(,%d)' % (
+                    item['weight'],
+                    item['in_mac_address'],
+                    ACROSS_SUBNET_TABLE))
+        else:
+            bucket = (
+                'bucket=weight=%d, mod_dl_dst:%s, resubmit(,%d)' % (
+                    item['weight'],
+                    item['mac_address'],
+                    ACROSS_SUBNET_TABLE))
+        buckets.append(bucket)
+
+    def _update_enc_actions(self, enc_actions, flow_rule,
+                            next_hop_tap_enabled):
+        # Add resubmit action to send to TAP table.
+        if next_hop_tap_enabled:
+            pp_corr = flow_rule['pp_corr']
+            pp_corr_tap_nh = flow_rule['next_hops'][0].get('pp_corr_tap_nh')
+            tap_nh_node_type = flow_rule['next_hops'][0].get(
+                'tap_nh_node_type', constants.DST_NODE)
+            group_action = enc_actions.split(',')[-1]
+            enc_actions = ""
+            if tap_nh_node_type == constants.SF_NODE:
+                if not pp_corr and pp_corr_tap_nh:
+                    if flow_rule.get('pc_corr', 'mpls') == 'mpls':
+                        mpls_act = self._build_push_mpls(flow_rule['nsp'],
+                                                         flow_rule['nsi'])
+                        enc_actions += mpls_act
+                    # enc_actions += group_action
+                enc_actions += group_action
+            else:
+                if flow_rule['pc_corr'] == 'mpls':
+                    # For DST Node
+                    if flow_rule['pp_corr']:
+                        enc_actions = ('pop_mpls:0x%04x,%s' % (
+                            constants.ETH_TYPE_IP, NORMAL_ACTION))
+                    else:
+                        enc_actions += NORMAL_ACTION
+            return enc_actions + RESUBMIT_TAP_TABLE
+        elif flow_rule.get('tap_enabled'):
+            return enc_actions + RESUBMIT_TAP_TABLE
+        return enc_actions
+
+    def _delete_across_subnet_table_flows(self, flowrule):
+        if not flowrule['next_hops']:
+            return
+        tap_enabled = flowrule['next_hops'][0].get('tap_enabled', False)
+        if tap_enabled:
+            egress_port = self.br_int.get_vif_port_by_id(flowrule['egress'])
+            for item in flowrule['next_hops']:
+                if flowrule['fwd_path']:
+                    self.br_int.delete_flows(
+                        table=TAP_CLASSIFIER_TABLE,
+                        dl_src=egress_port.vif_mac)
+                else:
+                    self.br_int.delete_flows(
+                        table=TAP_CLASSIFIER_TABLE,
+                        dl_src=egress_port.vif_mac)
+
+                if item['local_endpoint'] != self.local_ip:
+                    self._delete_tunnel_bridge_flows(flowrule,
+                                                     egress_port.vif_mac)
+        else:
+            for item in flowrule['next_hops']:
+                if flowrule['fwd_path']:
+                    self.br_int.delete_flows(
+                        table=ACROSS_SUBNET_TABLE,
+                        dl_dst=item['in_mac_address'])
+                else:
+                    self.br_int.delete_flows(
+                        table=ACROSS_SUBNET_TABLE,
+                        dl_dst=item['mac_address'])
+
+    def _add_tap_ingress_flow(self, flowrule, vif_port, vlan):
+        match_field = self._build_tap_ingress_match_field_sfc_mpls(
+            flowrule, vif_port, vlan)
+        actions = ('strip_vlan, pop_mpls:0x%04x,output:%s'
+                   % (constants.ETH_TYPE_MPLS,
+                      vif_port.ofport))
+        match_field['actions'] = actions
+        match_field.pop('dl_dst', None)
+        match_field.update(dl_src=flowrule['mac_address'])
+        self.br_int.add_flow(**match_field)
+
+    def _build_tap_ingress_match_field_sfc_mpls(self, flowrule, vif_port,
+                                                vlan):
+        match_field = self._build_ingress_common_match_field(vif_port, vlan)
+        match_field['eth_type'] = constants.ETH_TYPE_MPLS
+        match_field['mpls_label'] = flowrule['nsp'] << 8 | flowrule['nsi']
+        return match_field
+
+    def _delete_tunnel_bridge_flows(self, flowrule, src_mac):
+        match_info = {'in_port': self.patch_int_ofport,
+                      'dl_src': src_mac}
+        # Use Tap 'nsi'
+        flowrule_copy = flowrule.copy()
+        flowrule_copy['nsi'], flowrule_copy['nsp'] = (
+            flowrule['next_hops'][0]['nsi'], flowrule['next_hops'][0]['nsp'])
+        self._build_classification_match_sfc_mpls(flowrule_copy, match_info)
+        self.br_tun.delete_flows(table=0,
+                                 **match_info)
+        self.br_tun.delete_flows(table=TAP_TUNNEL_OUTPUT_TABLE,
+                                 **match_info)
+
+    def _clear_sfc_flow_on_tun_br(self):
+        self.br_tun.delete_flows(table=0, eth_type=constants.ETH_TYPE_MPLS)
+        self.br_tun.delete_flows(table=TAP_TUNNEL_OUTPUT_TABLE)
+
+    def _delete_flows_nsh(self, flowrule, vif_port):
         self.br_int.delete_flows(
             table=INGRESS_TABLE,
-            dl_type=constants.ETH_TYPE_MPLS,
+            eth_type=constants.ETH_TYPE_NSH,
             dl_dst=vif_port.vif_mac,
-            mpls_label=flowrule['nsp'] << 8 | flowrule['nsi'] + 1
+            nsh_mdtype=1,
+            nsh_spi=flowrule['nsp'],
+            nsh_si=flowrule['nsi'] + 1
         )
