@@ -13,20 +13,18 @@
 # limitations under the License.
 
 import abc
-import six
 
-from oslo_config import cfg
-from oslo_log import log as logging
-import oslo_messaging
-
-from neutron.agent.l2 import l2_agent_extension
 from neutron.agent import rpc as agent_rpc
 from neutron.common import rpc as n_rpc
 from neutron.common import topics
 from neutron import manager
+from neutron_lib.agent import l2_extension
+from oslo_config import cfg
+from oslo_log import log as logging
+import oslo_messaging
+import six
 
 from networking_sfc.services.sfc.drivers.ovs import rpc_topics as sfc_topics
-from networking_sfc._i18n import _LE, _LI
 
 LOG = logging.getLogger(__name__)
 
@@ -71,7 +69,7 @@ class SfcAgentDriver(object):
         """Delete a flow rule in driver."""
 
 
-class SfcAgentExtension(l2_agent_extension.L2AgentExtension):
+class SfcAgentExtension(l2_extension.L2AgentExtension):
 
     def initialize(self, connection, driver_type):
         """Initialize agent extension."""
@@ -108,7 +106,7 @@ class SfcAgentExtension(l2_agent_extension.L2AgentExtension):
                         flow, flowrule_status)
         except Exception as e:
             LOG.exception(e)
-            LOG.error(_LE("SFC L2 extension handle_port failed"))
+            LOG.error("SFC L2 extension handle_port failed")
             resync = True
 
         if flowrule_status:
@@ -121,13 +119,13 @@ class SfcAgentExtension(l2_agent_extension.L2AgentExtension):
         """Handle agent SFC extension port delete."""
         port_id = port['port_id']
         resync = False
-        LOG.info(_LI("a device %s is removed"), port_id)
+        LOG.info("a device %s is removed", port_id)
         try:
             self._delete_ports_flowrules_by_id(context, port_id)
         except Exception as e:
             LOG.exception(e)
             LOG.error(
-                _LE("delete port flow rule failed for %(port_id)s"),
+                "delete port flow rule failed for %(port_id)s",
                 {'port_id': port_id}
             )
             resync = True
@@ -146,7 +144,7 @@ class SfcAgentExtension(l2_agent_extension.L2AgentExtension):
                     flowrules, flowrule_status)
         except Exception as e:
             LOG.exception(e)
-            LOG.error(_LE("update_flow_rules failed"))
+            LOG.error("update_flow_rules failed")
 
         if flowrule_status:
             self.sfc_plugin_rpc.update_flowrules_status(
@@ -162,7 +160,7 @@ class SfcAgentExtension(l2_agent_extension.L2AgentExtension):
                     flowrules, flowrule_status)
         except Exception as e:
             LOG.exception(e)
-            LOG.error(_LE("delete_flow_rules failed"))
+            LOG.error("delete_flow_rules failed")
 
         if flowrule_status:
             self.sfc_plugin_rpc.update_flowrules_status(
@@ -204,7 +202,7 @@ class SfcAgentExtension(l2_agent_extension.L2AgentExtension):
                 "_delete_ports_flowrules_by_id received, count= %s", count)
         except Exception as e:
             LOG.exception(e)
-            LOG.error(_LE("delete_port_id_flows failed"))
+            LOG.error("delete_port_id_flows failed")
         if flowrule_status:
             self.sfc_plugin_rpc.update_flowrules_status(
                 context, flowrule_status)
