@@ -1,4 +1,4 @@
-# Copyright 2015 Futurewei. All rights reserved.
+# Copyright 2017 Futurewei. All rights reserved.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
 #    not use this file except in compliance with the License. You may obtain
@@ -13,20 +13,22 @@
 #    under the License.
 #
 
-import six
+from neutron.db import api as db_api
+from neutron.db import common_db_mixin
+from neutron_lib import context as n_context
+from neutron_lib.db import model_base
+from neutron_lib import exceptions as n_exc
+
 import sqlalchemy as sa
 from sqlalchemy import orm
 from sqlalchemy.orm import exc
 from sqlalchemy import sql
 
-from neutron_lib.db import model_base
-from neutron_lib import exceptions as n_exc
+import six
+
 from oslo_log import helpers as log_helpers
 from oslo_utils import uuidutils
 
-from neutron import context as n_context
-from neutron.db import common_db_mixin
-
 from networking_sfc._i18n import _
 
 
@@ -58,6 +60,7 @@ def singleton(class_):
         if class_ not in instances:
             instances[class_] = class_(*args, **kwargs)
         return instances[class_]
+
     return getinstance
 
 
@@ -67,11 +70,11 @@ class IDAllocation(object):
         # Get the initial range from conf file.
         conf_obj = {'group': [1, 255], 'portchain': [256, 65536]}
         self.conf_obj = conf_obj
-        self.session = context.session
+        self.context = context
 
     @log_helpers.log_method_call
     def assign_intid(self, type_, uuid):
-        query = self.session.query(UuidIntidAssoc).filter_by(
+        query = self.context.session.query(UuidIntidAssoc).filter_by(
             type_=type_).order_by(UuidIntidAssoc.intid)
 
         allocated_int_ids = {obj.intid for obj in query.all()}
@@ -81,23 +84,21 @@ class IDAllocation(object):
         start, end = self.conf_obj[type_][0], self.conf_obj[type_][1] + 1
         for init_id in six.moves.range(start, end):
             if init_id not in allocated_int_ids:
-                with self.session.begin(subtransactions=True):
+                with db_api.context_manager.writer.using(self.context):
                     uuid_intid = UuidIntidAssoc(
                         uuid, init_id, type_)
-                    self.session.add(uuid_intid)
+                    self.context.session.add(uuid_intid)
                 return init_id
-        else:
-            return None
+        return None
 
     @log_helpers.log_method_call
     def get_intid_by_uuid(self, type_, uuid):
 
-        query_obj = self.session.query(UuidIntidAssoc).filter_by(
+        query_obj = self.context.session.query(UuidIntidAssoc).filter_by(
             type_=type_, uuid=uuid).first()
         if query_obj:
             return query_obj.intid
-        else:
-            return None
+        return None
 
     @log_helpers.log_method_call
     def release_intid(self, type_, intid):
@@ -106,8 +107,8 @@ class IDAllocation(object):
         @param: type_: str
         @param: intid: int
         """
-        with self.session.begin(subtransactions=True):
-            query_obj = self.session.query(UuidIntidAssoc).filter_by(
+        with db_api.context_manager.writer.using(self.context):
+            query_obj = self.context.session.query(UuidIntidAssoc).filter_by(
                 intid=intid, type_=type_).first()
 
             if query_obj:
@@ -138,6 +139,7 @@ class PortPairDetail(model_base.BASEV2, model_base.HasId,
     ingress = sa.Column(sa.String(36), nullable=True)
     egress = sa.Column(sa.String(36), nullable=True)
     host_id = sa.Column(sa.String(255), nullable=False)
+    in_mac_address = sa.Column(sa.String(32))
     mac_address = sa.Column(sa.String(32), nullable=False)
     network_type = sa.Column(sa.String(8))
     segment_id = sa.Column(sa.Integer)
@@ -146,6 +148,7 @@ class PortPairDetail(model_base.BASEV2, model_base.HasId,
                                   backref='port_pair_detail',
                                   lazy="joined",
                                   cascade='all,delete')
+    correlation = sa.Column(sa.String(255), nullable=True)
 
 
 class PathNode(model_base.BASEV2, model_base.HasId, model_base.HasProject):
@@ -163,6 +166,15 @@ class PathNode(model_base.BASEV2, model_base.HasId, model_base.HasProject):
                                         cascade='all,delete')
     next_group_id = sa.Column(sa.Integer)
     next_hop = sa.Column(sa.String(512))
+    fwd_path = sa.Column(sa.Boolean(),
+                         nullable=False)
+    ppg_n_tuple_mapping = sa.Column(sa.String(1024), nullable=True)
+    tap_enabled = sa.Column(sa.Boolean(),
+                            nullable=False,
+                            server_default=sa.sql.false())
+    previous_node_id = sa.Column(
+        sa.String(36),
+        sa.ForeignKey('sfc_path_nodes.id', ondelete='SET NULL'))
 
 
 class OVSSfcDriverDB(common_db_mixin.CommonDbMixin):
@@ -171,7 +183,7 @@ class OVSSfcDriverDB(common_db_mixin.CommonDbMixin):
 
     def _make_pathnode_dict(self, node, fields=None):
         res = {'id': node['id'],
-               'tenant_id': node['tenant_id'],
+               'project_id': node['project_id'],
                'node_type': node['node_type'],
                'nsp': node['nsp'],
                'nsi': node['nsi'],
@@ -181,24 +193,30 @@ class OVSSfcDriverDB(common_db_mixin.CommonDbMixin):
                'status': node['status'],
                'portpair_details': [pair_detail['portpair_id']
                                     for pair_detail in node['portpair_details']
-                                    ]
+                                    ],
+               'fwd_path': node['fwd_path'],
+               'ppg_n_tuple_mapping': node['ppg_n_tuple_mapping'],
+               'tap_enabled': node['tap_enabled'],
+               'previous_node_id': node['previous_node_id']
                }
 
         return self._fields(res, fields)
 
     def _make_port_detail_dict(self, port, fields=None):
         res = {'id': port['id'],
-               'tenant_id': port['tenant_id'],
+               'project_id': port['project_id'],
                'host_id': port['host_id'],
                'ingress': port.get('ingress', None),
                'egress': port.get('egress', None),
                'segment_id': port['segment_id'],
                'local_endpoint': port['local_endpoint'],
                'mac_address': port['mac_address'],
+               'in_mac_address': port['in_mac_address'],
                'network_type': port['network_type'],
                'path_nodes': [{'pathnode_id': node['pathnode_id'],
                                'weight': node['weight']}
-                              for node in port['path_nodes']]
+                              for node in port['path_nodes']],
+               'correlation': port['correlation']
                }
 
         return self._fields(res, fields)
@@ -218,15 +236,15 @@ class OVSSfcDriverDB(common_db_mixin.CommonDbMixin):
             raise NodeNotFound(node_id=id)
         return node
 
-    def _get_port_detail(self, id):
+    def _get_port_pair_detail(self, id):
         try:
             port = self._get_by_id(self.admin_context, PortPairDetail, id)
         except exc.NoResultFound:
             raise PortPairDetailNotFound(port_id=id)
         return port
 
-    def create_port_detail(self, port):
-        with self.admin_context.session.begin(subtransactions=True):
+    def create_port_pair_detail(self, port):
+        with db_api.context_manager.writer.using(self.admin_context):
             args = self._filter_non_model_columns(port, PortPairDetail)
             args['id'] = uuidutils.generate_uuid()
             port_obj = PortPairDetail(**args)
@@ -234,7 +252,7 @@ class OVSSfcDriverDB(common_db_mixin.CommonDbMixin):
             return self._make_port_detail_dict(port_obj)
 
     def create_path_node(self, node):
-        with self.admin_context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(self.admin_context):
             args = self._filter_non_model_columns(node, PathNode)
             args['id'] = uuidutils.generate_uuid()
             node_obj = PathNode(**args)
@@ -242,22 +260,22 @@ class OVSSfcDriverDB(common_db_mixin.CommonDbMixin):
             return self._make_pathnode_dict(node_obj)
 
     def create_pathport_assoc(self, assoc):
-        with self.admin_context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(self.admin_context):
             args = self._filter_non_model_columns(assoc, PathPortAssoc)
             assoc_obj = PathPortAssoc(**args)
             self.admin_context.session.add(assoc_obj)
             return self._make_pathport_assoc_dict(assoc_obj)
 
     def delete_pathport_assoc(self, pathnode_id, portdetail_id):
-        with self.admin_context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(self.admin_context):
             self.admin_context.session.query(PathPortAssoc).filter_by(
                 pathnode_id=pathnode_id,
                 portpair_id=portdetail_id).delete()
 
     def update_port_detail(self, id, port):
-        with self.admin_context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(self.admin_context):
             port_obj = self._get_port_detail(id)
-            for key, value in six.iteritems(port):
+            for key, value in port.items():
                 if key == 'path_nodes':
                     pns = []
                     for pn in value:
@@ -283,13 +301,12 @@ class OVSSfcDriverDB(common_db_mixin.CommonDbMixin):
             return self._make_port_detail_dict(port_obj)
 
     def update_path_node(self, id, node):
-        with self.admin_context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(self.admin_context):
             node_obj = self._get_path_node(id)
-            for key, value in six.iteritems(node):
+            for key, value in node.items():
                 if key == 'portpair_details':
                     pds = []
                     for pd_id in value:
-                        self._get_port_detail(pd_id)
                         query = self._model_query(
                             self.admin_context, PathPortAssoc)
                         pd_association = query.filter_by(
@@ -305,25 +322,26 @@ class OVSSfcDriverDB(common_db_mixin.CommonDbMixin):
                     node_obj[key] = pds
                 else:
                     node_obj[key] = value
+
             return self._make_pathnode_dict(node_obj)
 
-    def delete_port_detail(self, id):
-        with self.admin_context.session.begin(subtransactions=True):
-            port_obj = self._get_port_detail(id)
+    def delete_port_pair_detail(self, id):
+        with db_api.context_manager.writer.using(self.admin_context):
+            port_obj = self._get_port_pair_detail(id)
             self.admin_context.session.delete(port_obj)
 
     def delete_path_node(self, id):
-        with self.admin_context.session.begin(subtransactions=True):
+        with db_api.context_manager.writer.using(self.admin_context):
             node_obj = self._get_path_node(id)
             self.admin_context.session.delete(node_obj)
 
     def get_port_detail(self, id):
-        with self.admin_context.session.begin(subtransactions=True):
-            port_obj = self._get_port_detail(id)
+        with db_api.context_manager.reader.using(self.admin_context):
+            port_obj = self._get_port_pair_detail(id)
             return self._make_port_detail_dict(port_obj)
 
     def get_port_detail_without_exception(self, id):
-        with self.admin_context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(self.admin_context):
             try:
                 port = self._get_by_id(
                     self.admin_context, PortPairDetail, id)
@@ -332,43 +350,30 @@ class OVSSfcDriverDB(common_db_mixin.CommonDbMixin):
             return self._make_port_detail_dict(port)
 
     def get_path_node(self, id):
-        with self.admin_context.session.begin(subtransactions=True):
+        with db_api.context_manager.reader.using(self.admin_context):
             node_obj = self._get_path_node(id)
         return self._make_pathnode_dict(node_obj)
 
-    def get_path_nodes_by_filter(self, filters=None, fields=None,
-                                 sorts=None, limit=None, marker=None,
-                                 page_reverse=False):
-        with self.admin_context.session.begin(subtransactions=True):
-            qry = self._get_path_nodes_by_filter(
-                filters, fields, sorts, limit,
-                marker, page_reverse
-            )
+    def get_path_nodes_by_filter(self, filters=None):
+        with db_api.context_manager.reader.using(self.admin_context):
+            qry = self._get_path_nodes_by_filter(filters)
             all_items = qry.all()
             if all_items:
                 return [self._make_pathnode_dict(item) for item in all_items]
-
         return None
 
-    def get_path_node_by_filter(self, filters=None, fields=None,
-                                sorts=None, limit=None, marker=None,
-                                page_reverse=False):
-        with self.admin_context.session.begin(subtransactions=True):
-            qry = self._get_path_nodes_by_filter(
-                filters, fields, sorts, limit,
-                marker, page_reverse)
+    def get_path_node_by_filter(self, filters=None):
+        with db_api.context_manager.reader.using(self.admin_context):
+            qry = self._get_path_nodes_by_filter(filters)
             first = qry.first()
             if first:
                 return self._make_pathnode_dict(first)
-
         return None
 
-    def _get_path_nodes_by_filter(self, filters=None, fields=None,
-                                  sorts=None, limit=None, marker=None,
-                                  page_reverse=False):
+    def _get_path_nodes_by_filter(self, filters=None):
         qry = self.admin_context.session.query(PathNode)
         if filters:
-            for key, value in six.iteritems(filters):
+            for key, value in filters.items():
                 column = getattr(PathNode, key, None)
                 if column:
                     if not value:
@@ -377,46 +382,31 @@ class OVSSfcDriverDB(common_db_mixin.CommonDbMixin):
                         qry = qry.filter(column == value)
         return qry
 
-    def get_port_details_by_filter(self, filters=None, fields=None,
-                                   sorts=None, limit=None, marker=None,
-                                   page_reverse=False):
-        with self.admin_context.session.begin(subtransactions=True):
-            qry = self._get_port_details_by_filter(
-                filters, fields, sorts, limit,
-                marker, page_reverse)
+    def get_port_details_by_filter(self, filters=None):
+        with db_api.context_manager.reader.using(self.admin_context):
+            qry = self._get_port_details_by_filter(filters)
             all_items = qry.all()
             if all_items:
-                return [
-                    self._make_port_detail_dict(item)
-                    for item in all_items
-                ]
-
+                return [self._make_port_detail_dict(item)
+                        for item in all_items]
         return None
 
-    def get_port_detail_by_filter(self, filters=None, fields=None,
-                                  sorts=None, limit=None, marker=None,
-                                  page_reverse=False):
-        with self.admin_context.session.begin(subtransactions=True):
-            qry = self._get_port_details_by_filter(
-                filters, fields, sorts, limit,
-                marker, page_reverse)
+    def get_port_detail_by_filter(self, filters=None):
+        with db_api.context_manager.reader.using(self.admin_context):
+            qry = self._get_port_details_by_filter(filters)
             first = qry.first()
             if first:
                 return self._make_port_detail_dict(first)
-
         return None
 
-    def _get_port_details_by_filter(self, filters=None, fields=None,
-                                    sorts=None, limit=None, marker=None,
-                                    page_reverse=False):
+    def _get_port_details_by_filter(self, filters=None):
         qry = self.admin_context.session.query(PortPairDetail)
         if filters:
-            for key, value in six.iteritems(filters):
+            for key, value in filters.items():
                 column = getattr(PortPairDetail, key, None)
                 if column:
                     if not value:
                         qry = qry.filter(sql.false())
                     else:
                         qry = qry.filter(column == value)
-
         return qry
