@@ -1,4 +1,5 @@
 # Copyright 2016 Futurewei. All rights reserved.
+# Copyright 2017 Intel Corporation.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
 #    not use this file except in compliance with the License. You may obtain
@@ -16,9 +17,9 @@ import socket
 
 import netaddr
 from tempest.api.network import base
+from tempest.common import utils
 from tempest.lib.common.utils import data_utils
 from tempest.lib import exceptions as lib_exc
-from tempest import test
 
 from networking_sfc.tests.tempest_plugin.tests import flowclassifier_client
 from networking_sfc.tests.tempest_plugin.tests import sfc_client
@@ -31,7 +32,7 @@ class BaseFlowClassifierTest(
     @classmethod
     def resource_setup(cls):
         super(BaseFlowClassifierTest, cls).resource_setup()
-        if not test.is_extension_enabled('flow_classifier', 'network'):
+        if not utils.is_extension_enabled('flow_classifier', 'network'):
             msg = "FlowClassifier Extension not enabled."
             raise cls.skipException(msg)
         cls.network = cls.create_network()
@@ -40,7 +41,7 @@ class BaseFlowClassifierTest(
 
     @classmethod
     def resource_cleanup(cls):
-        if not test.is_extension_enabled('flow_classifier', 'network'):
+        if not utils.is_extension_enabled('flow_classifier', 'network'):
             msg = "FlowClassifier Extension not enabled."
             raise cls.skipException(msg)
         super(BaseFlowClassifierTest, cls).resource_cleanup()
@@ -94,13 +95,13 @@ class BaseSfcTest(
     @classmethod
     def resource_setup(cls):
         super(BaseSfcTest, cls).resource_setup()
-        if not test.is_extension_enabled('sfc', 'network'):
+        if not utils.is_extension_enabled('sfc', 'network'):
             msg = "Sfc Extension not enabled."
             raise cls.skipException(msg)
 
     @classmethod
     def resource_cleanup(cls):
-        if not test.is_extension_enabled('sfc', 'network'):
+        if not utils.is_extension_enabled('sfc', 'network'):
             msg = "Sfc Extension not enabled."
             raise cls.skipException(msg)
         super(BaseSfcTest, cls).resource_cleanup()
@@ -174,3 +175,20 @@ class BaseSfcTest(
         body = self.portchain_client.list_port_chains()
         pc_list = body['port_chains']
         self.assertNotIn(pc_id, [n['id'] for n in pc_list])
+
+    def _try_create_service_graph(self, **kwargs):
+        graph = self.create_service_graph(
+            **kwargs)
+        self.addCleanup(self._try_delete_service_graph, graph['id'])
+        return graph
+
+    def _try_delete_service_graph(self, graph_id):
+        # delete Service Graph, if it exists
+        try:
+            self.sfcgraph_client.delete_service_graph(graph_id)
+        # if Service Graph is not found, this means it was deleted
+        except lib_exc.NotFound:
+            pass
+        body = self.sfcgraph_client.list_service_graphs()
+        graph_list = body['service_graphs']
+        self.assertNotIn(graph_id, [n['id'] for n in graph_list])
