@@ -14,17 +14,16 @@
 
 import time
 
-from tempest.common import waiters
+from oslo_log import log
+from tempest.common import utils
 from tempest import config
 from tempest.lib.common.utils import test_utils
 from tempest.lib import decorators
-from tempest import test
-import testscenarios
 
 from networking_sfc.tests.tempest_plugin.tests.scenario import base
 
 CONF = config.CONF
-load_tests = testscenarios.load_tests_apply_scenarios
+LOG = log.getLogger(__name__)
 
 
 class TestSfc(base.SfcScenarioTest):
@@ -38,7 +37,7 @@ class TestSfc(base.SfcScenarioTest):
             raise cls.skipException(msg)
         required_exts = ['sfc', 'flow_classifier']
         for ext in required_exts:
-            if not test.is_extension_enabled(ext, 'network'):
+            if not utils.is_extension_enabled(ext, 'network'):
                 msg = "%s Extension not enabled." % ext
                 raise cls.skipException(msg)
 
@@ -57,6 +56,15 @@ class TestSfc(base.SfcScenarioTest):
 
     def setUp(self):
         super(TestSfc, self).setUp()
+
+        self.multi_node = CONF.compute.min_compute_nodes > 1
+        if self.multi_node:
+            LOG.info("Running test on multi node")
+        else:
+            LOG.info("Running test on single node")
+        # Save servers UUIDs
+        self.servers = []
+
         self.ssh_user = CONF.validation.image_ssh_user
         self.keypair = self.create_keypair()
         self._setup_security_group()
@@ -82,9 +90,9 @@ class TestSfc(base.SfcScenarioTest):
             self._get_server_port_id_and_ip4(server))
         return floating_ip, port_id, fixed_ip
 
-    def _create_floating_ip(self, server):
+    def _create_floating_ip(self, server, client=None):
         floating_ip = self.create_floating_ip(
-            server)
+            server, client=client)
         self.check_floating_ip_status(floating_ip, 'ACTIVE')
         floating_ip_addr = floating_ip['floating_ip_address']
         self.check_public_network_connectivity(
@@ -147,12 +155,12 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21a8')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_create_port_chain(self):
         self._create_port_chain_helper(False)
 
     @decorators.idempotent_id('35927961-1904-4a6b-9d08-ad819f1cf812')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_create_port_chain_symmetric(self):
         self._create_port_chain_helper(True)
 
@@ -233,12 +241,12 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21a9')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_create_port_chain_multi_flow_classifiers(self):
         self._create_port_chain_multi_fc_helper(False)
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21b1')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_create_port_chain_multi_flow_classifiers_symmetric(self):
         self._create_port_chain_multi_fc_helper(True)
 
@@ -297,12 +305,12 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21aa')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_create_port_chain_multi_port_pairs(self):
         self._create_port_chain_multi_port_pairs_helper(False)
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f869be1e21ad')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_create_port_chain_multi_port_pairs_symmetric(self):
         self._create_port_chain_multi_port_pairs_helper(True)
 
@@ -364,17 +372,17 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21ab')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_create_port_chain_multi_port_pair_groups(self):
         self._create_port_chain_multi_ppg_helper(False)
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21b0')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_create_port_chain_multi_port_pair_groups_symmetric(self):
         self._create_port_chain_multi_ppg_helper(True)
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e22ab')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_create_multi_port_chain(self):
         (
             server1_floating_ip, server1_port_id, server1_fixed_ip
@@ -445,7 +453,7 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21ac')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_update_port_chain_add_flow_classifiers(self):
         (
             server1_floating_ip, server1_port_id, server1_fixed_ip
@@ -509,7 +517,7 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21ad')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_update_port_chain_remove_flow_classifiers(self):
         (
             server1_floating_ip, server1_port_id, server1_fixed_ip
@@ -573,7 +581,7 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21ae')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_update_port_chain_replace_flow_classifiers(self):
         (
             server1_floating_ip, server1_port_id, server1_fixed_ip
@@ -637,7 +645,7 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21af')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_update_port_chain_add_port_pair_groups(self):
         (
             server1_floating_ip, server1_port_id, server1_fixed_ip
@@ -697,7 +705,7 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21bf')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_update_port_chain_remove_port_pair_groups(self):
         (
             server1_floating_ip, server1_port_id, server1_fixed_ip
@@ -759,7 +767,7 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21be')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_update_port_chain_replace_port_pair_groups(self):
         (
             server1_floating_ip, server1_port_id, server1_fixed_ip
@@ -821,7 +829,7 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21bc')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_update_port_chain_replace_port_pair_groups_flow_classifiers(self):
         (
             server1_floating_ip, server1_port_id, server1_fixed_ip
@@ -899,7 +907,7 @@ class TestSfc(base.SfcScenarioTest):
         time.sleep(10)
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21bb')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_update_port_pair_group_add_port_pairs(self):
         (
             server1_floating_ip, server1_port_id, server1_fixed_ip
@@ -953,7 +961,7 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21ba')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_update_port_pair_group_remove_port_pairs(self):
         (
             server1_floating_ip, server1_port_id, server1_fixed_ip
@@ -1004,7 +1012,7 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21b9')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_update_port_pair_group_replace_port_pairs(self):
         (
             server1_floating_ip, server1_port_id, server1_fixed_ip
@@ -1055,7 +1063,7 @@ class TestSfc(base.SfcScenarioTest):
             private_key=self.keypair['private_key'])
 
     @decorators.idempotent_id('f970f6b3-6541-47ac-a9ea-f769be1e21b9')
-    @test.services('compute', 'network')
+    @utils.services('compute', 'network')
     def test_multi_port_chains_update_port_pair_group_replace_port_pairs(
         self
     ):
@@ -1167,18 +1175,33 @@ class TestSfc(base.SfcScenarioTest):
 
     def _create_server(self, network):
         kwargs = {}
-        if self.keypair is not None:
-            kwargs['key_name'] = self.keypair['name']
-        if self.security_group is not None:
-            kwargs['security_groups'] = [{'name': self.security_group['name']}]
-        server = self.create_server(
+        if self.multi_node:
+            kwargs["scheduler_hints"] = {'different_host': self.servers}
+
+        inst = self.create_server(
             networks=[{'uuid': network['id']}],
+            key_name=self.keypair['name'],
+            security_groups=[{'name': self.security_group['name']}],
             wait_until='ACTIVE',
             **kwargs)
-        waiters.wait_for_server_status(self.servers_client,
-                                       server['id'], 'ACTIVE')
+
+        adm_get_server = self.os_admin.servers_client.show_server
+        server = adm_get_server(inst['id'])['server']
+
         self._check_tenant_network_connectivity(
             server, self.ssh_user, self.keypair['private_key'])
+
+        # Check server is on different node
+        if self.multi_node:
+            new_host = server["OS-EXT-SRV-ATTR:host"]
+            host_list = [adm_get_server(s)["server"]["OS-EXT-SRV-ATTR:host"]
+                         for s in self.servers]
+            self.assertNotIn(new_host, host_list,
+                             message="Failed to create servers on different "
+                                     "Compute nodes.")
+
+            self.servers.append(server["id"])
+
         return server
 
     def _add_router_interface(self, router_id, port_id):
