@@ -14,17 +14,16 @@
 
 import contextlib
 import logging
-import mock
-import webob.exc
 
+import mock
+from neutron.api import extensions as api_ext
+from neutron.common import config
+import neutron.extensions as nextensions
 from neutron_lib import constants as lib_const
 from oslo_config import cfg
 from oslo_utils import importutils
 from oslo_utils import uuidutils
-
-from neutron.api import extensions as api_ext
-from neutron.common import config
-import neutron.extensions as nextensions
+import webob.exc
 
 from networking_sfc.db import flowclassifier_db as fdb
 from networking_sfc import extensions
@@ -101,6 +100,18 @@ class FlowClassifierDbPluginTestCaseBase(base.BaseTestCase):
         }
         return expected_flow_classifier
 
+    def _assert_flow_classifiers_match_subsets(self, flow_classifiers,
+                                               subsets, sort_key=None):
+        # Sort both lists
+        if sort_key:
+            flow_classifiers.sort(key=lambda fc: fc[sort_key])
+            subsets.sort(key=lambda fc: fc[sort_key])
+        for fc, subset in zip(flow_classifiers, subsets):
+            # Get matching items from subset
+            sub_fc = dict([(k, fc[k]) for k
+                           in subset.keys() if k in fc.keys()])
+            self.assertEqual(subset, sub_fc)
+
     def _test_create_flow_classifier(
         self, flow_classifier, expected_flow_classifier=None
     ):
