@@ -1,4 +1,4 @@
-# Copyright 2015 Futurewei. All rights reserved.
+# Copyright 2017 Futurewei. All rights reserved.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
 #    not use this file except in compliance with the License. You may obtain
@@ -13,19 +13,18 @@
 #    under the License.
 
 import copy
-import mock
-from webob import exc
-import webtest
 
-from oslo_utils import uuidutils
+import mock
 
 from neutron.api.v2 import resource as api_res_log
-from neutron.common import config as cfg
 from neutron import manager
 from neutron.notifiers import nova as nova_log
-
 from neutron.tests.unit.api.v2 import test_base as test_api_v2
 from neutron.tests.unit.extensions import base as test_api_v2_extension
+from oslo_config import cfg
+from oslo_utils import uuidutils
+from webob import exc
+import webtest
 
 from networking_sfc.extensions import sfc as sfc_ext
 
@@ -41,7 +40,7 @@ class SfcExtensionTestCase(test_api_v2_extension.ExtensionTestCase):
     fmt = 'json'
 
     def setUp(self):
-        self._mock_unncessary_looging()
+        self._mock_unnecessary_logging()
         super(SfcExtensionTestCase, self).setUp()
         self._setUpExtension(
             'networking_sfc.extensions.sfc.SfcPluginBase',
@@ -52,7 +51,7 @@ class SfcExtensionTestCase(test_api_v2_extension.ExtensionTestCase):
             plural_mappings={}
         )
 
-    def _mock_unncessary_looging(self):
+    def _mock_unnecessary_logging(self):
         mock_log_cfg_p = mock.patch.object(cfg, 'LOG')
         self.mock_log_cfg = mock_log_cfg_p.start()
 
@@ -115,7 +114,15 @@ class SfcExtensionTestCase(test_api_v2_extension.ExtensionTestCase):
     def test_create_port_chain_all_fields(self):
         self._test_create_port_chain(description='desc',
                                      name='test1',
-                                     chain_parameters={'correlation': 'mpls'},
+                                     chain_parameters={'symmetric': False,
+                                                       'correlation': 'mpls'},
+                                     flow_classifiers=[])
+
+    def test_create_port_chain_all_fields_with_symmetric(self):
+        self._test_create_port_chain(description='desc',
+                                     name='test1',
+                                     chain_parameters={'symmetric': True,
+                                                       'correlation': 'mpls'},
                                      flow_classifiers=[])
 
     def test_create_port_chain_none_chain_parameters(self):
@@ -188,7 +195,21 @@ class SfcExtensionTestCase(test_api_v2_extension.ExtensionTestCase):
         tenant_id = _uuid()
         data = {'port_chain': {
             'port_pair_groups': [_uuid()],
-            'chain_parameters': {'correlation': 'def'},
+            'chain_parameters': {'symmetric': False, 'correlation': 'def'},
+            'tenant_id': tenant_id, 'project_id': tenant_id
+        }}
+        self.assertRaises(
+            webtest.app.AppError,
+            self.api.post,
+            _get_path(PORT_CHAIN_PATH, fmt=self.fmt),
+            self.serialize(data),
+            content_type='application/%s' % self.fmt)
+
+    def test_create_port_chain_invalid_chain_parameters_symmetric(self):
+        tenant_id = _uuid()
+        data = {'port_chain': {
+            'port_pair_groups': [_uuid()],
+            'chain_parameters': {'symmetric': 'abc', 'correlation': 'mpls'},
             'tenant_id': tenant_id, 'project_id': tenant_id
         }}
         self.assertRaises(
@@ -325,7 +346,10 @@ class SfcExtensionTestCase(test_api_v2_extension.ExtensionTestCase):
             'project_id': port_pair_group['project_id'],
             'port_pair_group_parameters': port_pair_group.get(
                 'port_pair_group_parameters'
-            ) or {'lb_fields': []}
+            ) or {'lb_fields': [],
+                  'ppg_n_tuple_mapping': {'ingress_n_tuple': {},
+                                          'egress_n_tuple': {}}
+                  }
         }}
         if port_pair_group.get('group_id'):
             ret['port_pair_group']['group_id'] = port_pair_group['group_id']
@@ -471,6 +495,25 @@ class SfcExtensionTestCase(test_api_v2_extension.ExtensionTestCase):
             self.serialize(data),
             content_type='application/%s' % self.fmt)
 
+    def test_create_port_pair_group_invalid_ppg_n_tuple_mapping_key(self):
+        tenant_id = _uuid()
+        data = {'port_pair_group': {
+            'port_pairs': [_uuid()],
+            'port_pair_group_parameters': {
+                'ppg_n_tuple_mapping': {
+                    'ingress_n_tuple': {'sssource_ip_prefix': None},
+                    'egress_n_tuple': {'protool': None}
+                }
+            },
+            'tenant_id': tenant_id, 'project_id': tenant_id
+        }}
+        self.assertRaises(
+            webtest.app.AppError,
+            self.api.post,
+            _get_path(PORT_PAIR_GROUP_PATH, fmt=self.fmt),
+            self.serialize(data),
+            content_type='application/%s' % self.fmt)
+
     def test_create_port_pair_group_nonuuid_port_pairs(self):
         tenant_id = _uuid()
         data = {'port_pair_group': {
